(in-package "MORPHICL")

(xlib:define-keysym :cursor-home (xlib:keysym 255 80))
(xlib:define-keysym :cursor-left (xlib:keysym 255 81))
(xlib:define-keysym :cursor-up (xlib:keysym 255 82))
(xlib:define-keysym :cursor-right (xlib:keysym 255 83))
(xlib:define-keysym :cursor-down (xlib:keysym 255 84))
(xlib:define-keysym :page-up (xlib:keysym 255 85))
(xlib:define-keysym :page-down (xlib:keysym 255 86))
(xlib:define-keysym :cursor-end (xlib:keysym 255 87))

(defclass clx-port (port)
  ((clx-display  :type xlib:display  :accessor clx-port-display)
   (clx-screen   :type xlib:screen   :accessor clx-port-screen)
   (clx-window   :type xlib:window   :accessor clx-port-window)
   (clx-pixmap   :type xlib:pixmap   :accessor clx-port-pixmap)
   (clx-gc       :type xlib:gcontext :accessor clx-port-gc)
   (clx-fonts    :type hash-table    :accessor clx-port-fonts :initform (make-hash-table :test #'equal))))

(defclass clx-font (font)
  ((clx-font     :type xlib:font     :initarg :clx-font :accessor clx-font-font)))

(defun clx-port-error-handler (display error-name &key &allow-other-keys)
  (declare (ignore display))
  (format *error-output* "MorphiCL: clx-port error: ~A~%" error-name))

(defun clx-port-color (port color)
  (xlib:alloc-color (xlib:window-colormap (clx-port-window port))
                    (xlib:make-color :red (color-r color)
                                     :green (color-g color)
                                     :blue (color-b color))))

(defun make-clx-port (world &key (host "") (display 0) (screen 0))
  (let ((port (make-instance 'clx-port
                             :foreground-color +color-black+
                             :background-color +color-black+
                             :world world
                             :input-focus world)))
    (setf (clx-port-display port)
          (xlib:open-display host :display display))
    (setf (xlib:display-error-handler (clx-port-display port))
          #'clx-port-error-handler)
    (setf (clx-port-screen port)
          (nth screen 
               (xlib:display-roots (clx-port-display port))))
    (setf (clx-port-window port) 
          (xlib:create-window
            :parent (xlib:screen-root (clx-port-screen port))
            :x 0
            :y 0
            :width (round (xlib:screen-width (clx-port-screen port)) 2)
            :height (round (xlib:screen-height (clx-port-screen port)) 2)
            :event-mask (xlib:make-event-mask :exposure
                                              :key-press
                                              :key-release
                                              :button-press
                                              :button-release
                                              :structure-notify
                                              :pointer-motion)))
    (setf (clx-port-pixmap port)
          (xlib:create-pixmap
            :width (xlib:drawable-width (clx-port-window port))
            :height (xlib:drawable-height (clx-port-window port))
            :depth (xlib:drawable-depth (clx-port-window port))
            :drawable (clx-port-window port)))
    (setf (xlib:window-background (clx-port-window port)) 
          (clx-port-pixmap port))
    (xlib:map-window (clx-port-window port))
    (setf (clx-port-gc port)
          (xlib:create-gcontext
            :drawable (clx-port-window port)
            :foreground (xlib:screen-white-pixel (clx-port-screen port))
            :background (xlib:screen-black-pixel (clx-port-screen port))
            :line-style :solid
            :line-width 1))
    (setf (canvas-font port)
          (make-font port "fixed"))
    (setf (canvas-transformation port)
          (make-rectangle 0
                          0
                          (xlib:drawable-width (clx-port-window port))
                          (xlib:drawable-height (clx-port-window port))))
    (setf (canvas-clipping-rectangle port)
          (make-rectangle 0
                          0
                          (xlib:drawable-width (clx-port-window port))
                          (xlib:drawable-height (clx-port-window port))))
    port))
            
(defmethod port-close ((port clx-port))
  (maphash (lambda (name font)
             (declare (ignore name))
             (xlib:close-font (clx-font-font font)))
           (clx-port-fonts port))
  (xlib:free-gcontext (clx-port-gc port))
  (xlib:free-pixmap (clx-port-pixmap port))
  (xlib:destroy-window (clx-port-window port))
  (xlib:close-display (clx-port-display port)))

(defmethod port-read-event ((port clx-port) &key timeout)
  (xlib:event-case ((clx-port-display port) :timeout timeout :discard-p t)
    (:exposure (x y width height)
      (xlib:clear-area (clx-port-window port) 
                       :x x 
                       :y y 
                       :width width 
                       :height height)
      nil)
    (:configure-notify (width height)
      (when (or (/= width (xlib:drawable-width (clx-port-pixmap port)))
                  (/= height (xlib:drawable-height (clx-port-pixmap port))))
        (xlib:free-pixmap (clx-port-pixmap port))
        (setf (clx-port-pixmap port)
              (xlib:create-pixmap
                :width width
                :height height
                :depth (xlib:drawable-depth (clx-port-window port))
                :drawable (clx-port-window port)))
        (setf (xlib:window-background (clx-port-window port))
              (clx-port-pixmap port))
        (setf (canvas-transformation port)
              (make-rectangle 0 0 width height))
        (setf (canvas-clipping-rectangle port)
              (make-rectangle 0 0 width height))
        (make-instance 'display-resize-event
                       :port port
                       :position (make-point 0 0)
                       :bounds (make-rectangle 0 0 width height))))
    (:motion-notify (x y)
      (make-instance 'pointer-motion-event
                     :port port
                     :position (make-point x y)))
    (:button-press (code x y)
      (make-instance 'pointer-button-press-event
                     :port port
                     :position (make-point x y)
                     :button code))
    (:button-release (code x y)
      (make-instance 'pointer-button-release-event
                     :port port
                     :position (make-point x y)
                     :button code))
    (:key-press (code state x y)
      (make-instance 'key-press-event
                     :port port
                     :position (make-point x y)
                     :key (xlib:keycode->character (clx-port-display port) code state)))
    (:key-release (code state x y)
      (make-instance 'key-release-event
                     :port port
                     :position (make-point x y)
                     :key (xlib:keycode->character (clx-port-display port) code state)))))

(defmethod morph-full-draw :after ((morph world-morph) (port clx-port))
  (xlib:clear-area (clx-port-window port)))

(defmethod make-font ((port clx-port) name)
  (or (gethash name (clx-port-fonts port))
      (setf (gethash name (clx-port-fonts port))
            (make-instance 'clx-font
                           :canvas port
                           :name name
                           :clx-font 
                             (xlib:open-font (clx-port-display port) name)))))
 
(defmethod font-width ((font clx-font) &optional text &key (start 0) end)
  (if text
      (multiple-value-bind (width)
        (xlib:text-extents (clx-font-font font) text :start start :end end)
        width)
      (xlib:max-char-width (clx-font-font font))))

(defmethod font-height ((font clx-font) &optional text &key (start 0) end)
  (if text
      (multiple-value-bind (width ascent)
        (xlib:text-extents (clx-font-font font) text :start start :end end)
        (declare (ignore width))
        ascent)
      (xlib:max-char-ascent (clx-font-font font))))

(defmethod canvas-width ((port clx-port))
  (xlib:drawable-width (clx-port-window port)))

(defmethod canvas-height ((port clx-port))
  (xlib:drawable-height (clx-port-window port)))

(defmethod (setf canvas-foreground-color) :after ((color color) (port clx-port))
  (setf (xlib:gcontext-foreground (clx-port-gc port))
        (clx-port-color port color)))

(defmethod (setf canvas-background-color) :after ((color color) (port clx-port))
  (setf (xlib:gcontext-background (clx-port-gc port))
        (clx-port-color port color)))

(defmethod (setf canvas-font) :after ((font clx-font) (port clx-port))
  (setf (xlib:gcontext-font (clx-port-gc port)) (clx-font-font font)))

(defmethod (setf canvas-clipping-rectangle) :after ((rectangle rectangle) (port clx-port))
  (setf (xlib:gcontext-clip-mask (clx-port-gc port) :yx-banded)
        (list (rectangle-left rectangle)
              (rectangle-top rectangle)
              (rectangle-width rectangle) 
              (rectangle-height rectangle))))

(defmethod canvas-draw-line ((port clx-port) (start point) (end point))
  (xlib:draw-line (clx-port-pixmap port)
                  (clx-port-gc port)
                  (point-x start)
                  (point-y start)
                  (point-x end)
                  (point-y end)))

(defmethod canvas-draw-lines ((port clx-port) points)
  (xlib:draw-segments (clx-port-pixmap port)
                      (clx-port-gc port)
                      (loop for point in points
                        collect (point-x point)
                        collect (point-y point))))
  
(defmethod canvas-draw-point ((port clx-port) (point point))
  (xlib:draw-point (clx-port-pixmap port)
                   (clx-port-gc port)
                   (point-x point)
                   (point-y point)))

(defmethod canvas-draw-points ((port clx-port) points)
  (xlib:draw-points (clx-port-pixmap port)
                    (clx-port-gc port)
                    (loop for point in points
                      collect (point-x point)
                      collect (point-y point))))

(defmethod canvas-outline-polygon ((port clx-port) points)
  (xlib:draw-lines (clx-port-pixmap port)
                   (clx-port-gc port)
                   (loop for point in points
                     collect (point-x point) into new-points
                     collect (point-y point) into new-points
                     finally
                       (return (nconc new-points 
                                     (list (first new-points)
                                           (second new-points)))))))

(defmethod canvas-fill-polygon ((port clx-port) points)
  (xlib:draw-lines (clx-port-pixmap port)
                   (clx-port-gc port)
                   (loop for point in points
                     collect (point-x point) into new-points
                     collect (point-y point) into new-points
                     finally
                       (return (nconc new-points 
                                     (list (first new-points)
                                           (second new-points)))))
                   :fill-p t))

(defmethod canvas-outline-rectangle ((port clx-port) (rectangle rectangle))
  (xlib:draw-rectangle (clx-port-pixmap port)
                       (clx-port-gc port)
                       (rectangle-left rectangle)
                       (rectangle-top rectangle)
                       (rectangle-width rectangle)
                       (rectangle-height rectangle)))

(defmethod canvas-fill-rectangle ((port clx-port) (rectangle rectangle))
  (xlib:draw-rectangle (clx-port-pixmap port)
                       (clx-port-gc port)
                       (rectangle-left rectangle)
                       (rectangle-top rectangle)
                       (rectangle-width rectangle)
                       (rectangle-height rectangle)
                       t))

(defmethod canvas-outline-ellipse ((port clx-port) (rectangle rectangle) 
                                   &optional (start 0) (end (* 2 pi)))
  (xlib:draw-arc (clx-port-pixmap port)
                 (clx-port-gc port)
                 (rectangle-left rectangle)
                 (rectangle-top rectangle)
                 (rectangle-width rectangle)
                 (rectangle-height rectangle)
                 start
                 end))

(defmethod canvas-fill-ellipse ((port clx-port) (rectangle rectangle)
                                &optional (start 0) (end (* 2 pi)))
  (xlib:draw-arc (clx-port-pixmap port)
                 (clx-port-gc port)
                 (rectangle-left rectangle)
                 (rectangle-top rectangle)
                 (rectangle-width rectangle)
                 (rectangle-height rectangle)
                 start
                 end
                 t))

(defmethod canvas-draw-text ((port clx-port) text (offset point) &key (start 0) end)
  (xlib:draw-glyphs (clx-port-pixmap port)
                    (clx-port-gc port)
                    (point-x offset)
                    (+ (point-y offset)
                       (font-height (canvas-font port)))
                    text
                    :start start
                    :end end))

