(in-package "MORPHICL")

(defclass morph ()
  ((bounds          :type rectangle        :initarg :bounds          :accessor morph-bounds)
   (color           :type color            :initarg :color           :accessor morph-color)
   (highlight-color :type color            :initarg :highlight-color :accessor morph-highlight-color)
   (owner           :type (or morph null)  :initarg :owner           :accessor morph-owner)
   (submorphs       :type (list morph)     :initform '()             :accessor morph-submorphs)))

(defun morphp (object)
  (typep object 'morph))

(defgeneric copy-morph ((morph morph)))
(defgeneric deep-copy-morph ((morph morph)))
(defgeneric morph-bounds ((morph morph)))
(defgeneric (setf morph-bounds) ((morph morph) (rectangle rectangle)))
(defgeneric morph-color ((morph morph)))
(defgeneric (setf morph-color) ((morph morph) (color color)))
(defgeneric morph-highlight-color ((morph morph)))
(defgeneric (setf morph-highlight-color) ((morph morph) (color color)))
(defgeneric morph-owner ((morph morph)))
(defgeneric (setf morph-owner) ((morph morph) (owner morph)))
(defgeneric morph-submorphs ((morph morph)))
(defgeneric (setf morph-submorphs) (submorphs (morph morph)))

(defgeneric morph-find-submorph ((morph morph) (submorph morph)))
(defgeneric morph-find-submorph-if ((morph morph) predicate))
(defgeneric morph-find-submorph-at ((morph morph) (point point)))
(defgeneric morph-add-submorph ((morph morph) (submorph morph)))
(defgeneric morph-add-submorph-front ((morph morph) (submorph morph)))
(defgeneric morph-add-submorph-back ((morph morph) (submorph morph)))
(defgeneric morph-add-submorph-before ((morph morph) (submorph morph) (successor morph)))
(defgeneric morph-add-submorph-before-if ((morph morph) (submorph morph) predicate))
(defgeneric morph-add-submorph-after ((morph morph) (submorph morph) (predecessor morph)))
(defgeneric morph-add-submorph-after-if ((morph morph) (submorph morph) predicate))
(defgeneric morph-add-submorph-at ((morph morph) (submorph morph) position))
(defgeneric morph-remove-submorph ((morph morph) (submorph morph)))
(defgeneric morph-remove-submorph-if ((morph morph) predicate))
(defgeneric morph-remove-submorph-at ((morph morph) position))
(defgeneric morph-substitute-submorph ((morph morph) (new morph) (old morph)))

(defgeneric morph-draw ((morph morph) (canvas canvas)))
(defgeneric morph-full-draw ((morph morph) (canvas canvas)))
(defgeneric morph-handle-event ((morph morph) (event event)))
(defgeneric morph-full-handle-event ((morph morph) (event event)))

(defmethod copy-morph ((old-morph morph))
  (let ((new-morph (make-instance (class-of old-morph))))
    (dolist (slot (mop:class-slots (class-of old-morph)))
      (let ((slot-name (mop:slot-definition-name slot)))
        (case slot-name
          (owner)
          (submorphs)
          (bounds
            (setf (morph-bounds new-morph)
                  (copy-rectangle (morph-bounds old-morph))))
          (otherwise
             (setf (slot-value new-morph slot-name)
                   (slot-value old-morph slot-name))))))
    new-morph))

(defmethod deep-copy-morph ((old-morph morph))
  (let ((new-morph (copy-morph old-morph)))
    (dolist (submorph (morph-submorphs old-morph))
      (morph-add-submorph-front new-morph
        (deep-copy-morph submorph)))
    new-morph))

(defmethod morph-find-submorph-at ((morph morph) (point point))
  (let ((submorph
          (find-if
            (lambda (submorph)
              (and (>= (point-x point) 
                       (rectangle-left (morph-bounds submorph)))
                   (< (point-x point) 
                      (rectangle-right (morph-bounds submorph)))
                   (>= (point-y point) 
                       (rectangle-top (morph-bounds submorph)))
                   (< (point-y point) 
                      (rectangle-bottom (morph-bounds submorph)))))
            (morph-submorphs morph))))
    (cond
      (submorph
        (point-translate point
          (- (rectangle-left (morph-bounds submorph)))
          (- (rectangle-top (morph-bounds submorph))))
        (point-scale point
          (/ 1 (rectangle-width (morph-bounds submorph)))
          (/ 1 (rectangle-height (morph-bounds submorph))))
        (morph-find-submorph-at submorph point))
      (t morph))))

(defmethod morph-find-submorph ((morph morph) (submorph morph))
  (position submorph (morph-submorphs morph)))

(defmethod morph-find-submorph-if ((morph morph) predicate)
  (position-if predicate (morph-submorphs morph)))

(defmethod morph-add-submorph-at ((morph morph) (submorph morph) position)
  (setf (morph-owner submorph) morph)
  (if (or (<= position 0)
          (null (morph-submorphs morph)))
      (push submorph (morph-submorphs morph))
      (push submorph (rest (or (nthcdr (1- position) (morph-submorphs morph))
                               (last (morph-submorphs morph)))))))

(defmethod morph-add-submorph ((morph morph) (submorph morph))
  (morph-add-submorph-front morph submorph))

(defmethod morph-add-submorph-front ((morph morph) (submorph morph))
  (morph-add-submorph-at morph submorph (length (morph-submorphs morph))))

(defmethod morph-add-submorph-back ((morph morph) (submorph morph))
  (morph-add-submorph-at morph submorph 0))

(defmethod morph-add-submorph-before ((morph morph) (submorph morph) (successor morph))
  (morph-add-submorph-at morph submorph (position successor (morph-submorphs morph))))

(defmethod morph-add-submorph-before-if ((morph morph) (submorph morph) predicate)
  (morph-add-submorph-before morph
                             submorph
                             (find-if predicate (morph-submorphs morph))))

(defmethod morph-add-submorph-after ((morph morph) (submorph morph) (predecessor morph))
  (morph-add-submorph-at morph 
                         submorph
                         (let ((position (position predecessor (morph-submorphs morph))))
                           (if position
                               (1+ position)
                               0))))

(defmethod morph-add-submorph-after-if ((morph morph) (submorph morph) predicate)
  (morph-add-submorph-after morph
                            submorph
                            (find-if predicate (morph-submorphs morph))))

(defmethod morph-remove-submorph ((morph morph) (submorph morph))
  (setf (morph-owner submorph) nil)
  (setf (morph-submorphs morph)
        (delete submorph (morph-submorphs morph))))

(defmethod morph-remove-submorph-if ((morph morph) predicate)
  (dolist (submorph (morph-submorphs morph))
    (when (funcall predicate submorph)
      (morph-remove-submorph morph submorph)))
  (morph-submorphs morph))

(defmethod morph-remove-submorph-at ((morph morph) position)
  (morph-remove-submorph morph (nth position (morph-submorphs morph))))

(defmethod morph-substitute-submorph ((morph morph) (new morph) (old morph))
  (setf (morph-submorphs morph)
        (nsubst new old (morph-submorphs morph))))

(defmethod morph-draw ((morph morph) (canvas canvas))
  (canvas-fill-rectangle canvas
                         (canvas-clipping-rectangle canvas)))

(defmethod morph-full-draw ((morph morph) (canvas canvas))
  (with-canvas-options (canvas 
                        :foreground-color (morph-color morph)
                        :transformation
                          (canvas-transform-rectangle canvas
                            (copy-rectangle (morph-bounds morph)))
                        :clipping-rectangle 
                          (rectangle-intersect
                            (canvas-transform-rectangle canvas
                              (copy-rectangle (morph-bounds morph)))
                            (canvas-clipping-rectangle canvas)))
    (morph-draw morph canvas)
    (dolist (submorph (morph-submorphs morph))
      (morph-full-draw submorph canvas))))

(defmethod morph-handle-event ((morph morph) (event event))
  nil)

(defmethod morph-full-handle-event ((morph morph) (event event))
  (unless (or (morph-handle-event morph event)
              (null (morph-owner morph)))
    (point-translate 
      (point-scale (event-position event)
        (rectangle-width (morph-bounds morph))
        (rectangle-height (morph-bounds morph)))
      (rectangle-left (morph-bounds morph))
      (rectangle-right (morph-bounds morph)))
    (morph-full-handle-event (morph-owner morph) event)))

  
