(in-package "MORPHICL")

(defclass point ()
  ((x    :type number  :initarg :x    :accessor point-x)
   (y    :type number  :initarg :y    :accessor point-y)))

(defun pointp (object)
  (typep object 'point))

(defgeneric point-x ((point point)))
(defgeneric (setf point-x) (x (point point)))
(defgeneric point-y ((point point)))
(defgeneric (setf point-y) (y (point point)))
(defgeneric point-position ((point point)))
(defgeneric* (setf point-position) (x y (point point)))

(defgeneric point-distance ((a point) (b point)))
(defgeneric point-round ((point point)))
(defgeneric point-translate ((point point) x y))
(defgeneric point-scale ((point point) x y))
(defgeneric point+ ((a point) (b point)))
(defgeneric point- ((a point) (b point)))
(defgeneric point-min ((a point) (b point)))
(defgeneric point-max ((a point) (b point)))
(defgeneric point= ((a point) (b point)))
(defgeneric point/= ((a point) (b point)))
(defgeneric point< ((a point) (b point)))
(defgeneric point<= ((a point) (b point)))
(defgeneric point> ((a point) (b point)))
(defgeneric point>= ((a point) (b point)))

(defun make-point (x y)
  (make-instance 'point :x x :y y))

(defun copy-point (point)
  (make-instance 'point :x (point-x point) :y (point-y point)))

(defmethod point-position ((point point))
  (values (point-x point) (point-y point)))

(defmethod* (setf point-position) (x y (point point))
  (values (setf (point-x point) x)
          (setf (point-y point) y)))

(defmethod point-round ((point point))
  (setf (point-x point) (round (point-x point)))
  (setf (point-y point) (round (point-y point)))
  point)

(defmethod point-distance ((a point) (b point))
  (let ((delta-x (- (point-x a) (point-x b)))
        (delta-y (- (point-y a) (point-y b))))
    (sqrt (+ (* delta-x delta-x) (* delta-y delta-y)))))

(defmethod point-translate ((point point) x y)
  (incf (point-x point) x)
  (incf (point-y point) y)
  point)

(defmethod point-scale ((point point) x y)
  (setf (point-x point) (* x (point-x point)))
  (setf (point-y point) (* y (point-y point)))
  point)

(defmethod point+ ((a point) (b point))
  (make-instance 'point 
                 :x (+ (point-x a) (point-x b))
                 :y (+ (point-y a) (point-y b))))

(defmethod point- ((a point) (b point))
  (make-instance 'point
                 :x (- (point-x a) (point-x b))
                 :y (- (point-y a) (point-y b))))

(defmethod point-min ((a point) (b point))
  (make-instance 'point
                 :x (min (point-x a) (point-x b))
                 :y (min (point-y a) (point-y b))))

(defmethod point-max ((a point) (b point))
  (make-instance 'point
                 :x (max (point-x a) (point-x b))
                 :y (max (point-y a) (point-y b))))

(defmethod point= ((a point) (b point))
  (and (= (point-x a) (point-x b))
       (= (point-y a) (point-y b))))

(defmethod point/= ((a point) (b point))
  (or (/= (point-x a) (point-x b))
      (/= (point-y a) (point-y b))))

(defmethod point< ((a point) (b point))
  (and (< (point-x a) (point-x b))
       (< (point-y a) (point-y b))))

(defmethod point<= ((a point) (b point))
  (and (<= (point-x a) (point-x b))
       (<= (point-y a) (point-y b))))

(defmethod point> ((a point) (b point))
  (and (> (point-x a) (point-x b))
       (> (point-y a) (point-y b))))

(defmethod point>= ((a point) (b point))
  (and (>= (point-x a) (point-x b))
       (>= (point-y a) (point-y b))))

