(in-package "MORPHICL")

(defclass port (canvas)
  ((world         :type world-morph     :initarg :world       :accessor port-world)
   (unread-events :type (list event)    :initform '()         :accessor port-unread-events)
   (baggage       :type (list morph)    :initform '()         :accessor port-baggage)
   (input-focus   :type morph           :initarg :input-focus :accessor port-input-focus)))

(defvar *default-port* nil)

(defgeneric port-world ((port port)))
(defgeneric (setf port-world) ((world world-morph) (port port)))
(defgeneric port-input-focus ((port port)))
(defgeneric (setf port-input-focus) ((input-focus morph) (port port)))

(defgeneric port-read-event ((port port) &key timeout))
(defgeneric port-unread-event ((port port) (event event)))
(defgeneric port-peek-event ((port port)))
(defgeneric port-handle-event ((port port) (event event)))
(defgeneric port-listen ((port port)))
(defgeneric port-close ((port port)))

(defun portp (object)
  (typep object 'port))

(defmethod port-listen ((port port))
  (let ((old-port *default-port*))
    (setf *default-port* port)
    (morph-full-draw (port-world port) port)
    (loop for event = (port-read-event port)
          while event
      do 
        (port-handle-event port event)
        (loop for event = (port-read-event port :timeout 0)
              while event
          do (port-handle-event port event))
        (when (port-baggage port)
          (morph-full-draw (port-world port) port)))
    (shiftf *default-port* old-port)))

(defmethod port-read-event :around ((port port) &key timeout)
  (declare (ignore timeout))
  (if (port-unread-events port)
      (pop (port-unread-events port))
      (call-next-method)))

(defmethod port-unread-event ((port port) (event event))
  (push event (port-unread-events port)))

(defmethod port-peek-event ((port port))
  (let ((event (port-read-event port :timeout 0)))
    (when event
      (port-unread-event port event))
    event))

(defmethod port-handle-event ((port port) (event event))
  (typecase event
    ((or pointer-event keyboard-event)
       (point-scale (event-position event)
         (/ 1 (canvas-width port))
         (/ 1 (canvas-height port)))
       (dolist (baggage (port-baggage port))
         (rectangle-translate
           (morph-bounds baggage)
           (- (point-x (event-position event))
              (rectangle-left (morph-bounds baggage)))
           (- (point-y (event-position event))
              (rectangle-top (morph-bounds baggage)))))
       (morph-full-handle-event
         (morph-find-submorph-at (port-world port)
                                 (event-position event))
         event))
    ((or display-redraw-event display-resize-event)
       (morph-full-draw (port-world port) port))))


